const deepEqual = require('fast-deep-equal');
const querystring = require('querystring');
const xhr = require('request');

const produceURL = 'https://www.varsnap.com/api/snap/produce/';
const consumeURL = 'https://www.varsnap.com/api/snap/consume/';
const produceTrialURL = 'https://www.varsnap.com/api/trial/produce/';
const version = 'v0.4.5';

const configKeys = {
  varsnap: 'varsnap',
  env: 'env',
  producerToken: 'producerToken',
  consumerToken: 'consumerToken',
};
let Producers = [];
let Consumers = [];

class Util {
  static getConfig(config, configKey) {
    if (!config) {
      return '';
    }
    let val = config[configKey];
    if (!val) {
      return '';
    }
    val = val.toString().toLowerCase();
    return val;
  }

  static getSignature(targetFunc) {
    let signature = targetFunc.toString().split('{')[0];
    signature = signature.replace(/^\s+|\s+$/g, '');
    return 'javascript.' + version + '.' + signature;
  }

  static ajax(url, requestData) {
    const options = {
      method: 'POST',
      uri: url,
      body: querystring.stringify(requestData).toString('utf8'),
      headers: {'Content-Type': 'application/x-www-form-urlencoded'},
    };
    const ajaxPromise = new Promise((resolve, reject) => {
      xhr(options, (err, resp, body) => {
        if (err) {
          return reject(err);
        }
        const responseCode = parseInt(resp.statusCode / 100, 10);
        if (responseCode != 2) {
          return reject(new Error('Status: ' + resp.statusCode));
        }
        body = JSON.parse(body);
        return resolve(body);
      });
    });
    return ajaxPromise;
  }

  static getLogger() {
    if (Util.logger) {
      return Util.logger;
    }
    return console; // eslint-disable-line no-console
  }

  static log(data) {
    Util.getLogger().log(data);
  }

  static error(data) {
    Util.getLogger().error(data);
  }

  static resetProducers() {
    Producers = [];
  }

  static resetConsumers() {
    Consumers = [];
  }

  static getProducers() {
    return Producers;
  }

  static getConsumers() {
    return Consumers;
  }
}

class Producer {
  constructor(targetFunc, config) {
    this.targetFunc = targetFunc;
    this.config = config;
  }

  isEnabled() {
    if (Util.getConfig(this.config, configKeys.varsnap) !== 'true') {
      return false;
    }
    if (Util.getConfig(this.config, configKeys.env) !== 'production') {
      return false;
    }
    if (!Util.getConfig(this.config, configKeys.producerToken)) {
      return false;
    }
    return true;
  }

  static serialize(data) {
    return JSON.stringify(data);
  }

  produce(args, output) {
    if (!this.isEnabled()) {
      return;
    }
    Util.log('Sending call to VarSnap');
    const data = {
      producer_token: Util.getConfig(this.config, configKeys.producerToken),
      signature: Util.getSignature(this.targetFunc),
      inputs: Producer.serialize(args),
      prod_outputs: Producer.serialize(output),
    };
    Util.ajax(produceURL, data).catch((err) => {
      if(err.statusCode === 404) {
        Util.error('Cannot produce to VarSnap; check your producer token');
      } else {
        Util.error(err.error);
      }
    });
  }
}

class Consumer {
  constructor(targetFunc, config) {
    this.targetFunc = targetFunc;
    this.config = config;
  }

  isEnabled() {
    if (Util.getConfig(this.config, configKeys.varsnap) !== 'true') {
      return false;
    }
    if (Util.getConfig(this.config, configKeys.env) !== 'development') {
      return false;
    }
    if (!Util.getConfig(this.config, configKeys.consumerToken)) {
      return false;
    }
    return true;
  }

  static deserialize(data) {
    if (data === '') {
      return undefined;
    }
    return JSON.parse(data);
  }

  consumeOnce() {
    if (!this.isEnabled()) {
      return;
    }
    const data = {
      consumer_token: Util.getConfig(this.config, configKeys.consumerToken),
      signature: Util.getSignature(this.targetFunc),
    };
    return Util.ajax(consumeURL, data).then((response) => {
      if (!response || response.status !== 'ok') {
        return false;
      }
      let allMatches = true;
      for(let snap of response.results) {
        Util.log('Receiving call from VarSnap uuid: ' + snap['id']);
        let inputs = Consumer.deserialize(snap['inputs']);
        let prodOutputs = Consumer.deserialize(snap['prod_outputs']);
        let localOutputs = this.targetFunc(...inputs);
        let matches = deepEqual(prodOutputs, localOutputs);
        this.reportLog(inputs, prodOutputs, localOutputs, matches);
        this.reportCentral(snap, localOutputs, matches);
        matches = allMatches && matches;
      }
      return allMatches;
    }).catch((error) => {
      Util.error('Cannot contact VarSnap: ' + error);
      return false;
    });
  }

  reportCentral(snap, localOutputs, matches) {
    const data = {
      'consumer_token': Util.getConfig(this.config, configKeys.consumerToken),
      'snap_id': snap.id,
      'dev_outputs': localOutputs,
      'matches': matches,
    };
    return Util.ajax(produceTrialURL, data).then((response) => {
      if (!response || response.status !== 'ok') {
        return false;
      }
      const trialURL = response['trial_url'];
      Util.log('Trial URL:                     ' + trialURL);
      return true;
    }).catch((error) => {
      Util.error('Cannot upload test results to VarSnap: ' + error);
      return false;
    });
  }

  reportLog(inputs, prodOutputs, localOutputs, matches) {
    const functionName = Util.getSignature(this.targetFunc);
    Util.log('Function:                      ' + functionName);
    Util.log('Function input args:           ' + inputs);
    Util.log('Production function outputs:   ' + prodOutputs);
    Util.log('Your function outputs:         ' + localOutputs);
    Util.log('Matching outputs:              ' + matches);
    Util.log('');
  }
}

function varsnap(func) {
  if (varsnap.config.logger) {
    Util.logger = varsnap.config.logger;
  }
  const producer = new Producer(func, varsnap.config);
  const consumer = new Consumer(func, varsnap.config);

  const wrapped = function(...args) {
    let result = '';
    try {
      result = func(...args);
    }
    catch(err) {
      producer.produce(args, err);
      throw(err);
    }
    producer.produce(args, result);
    return result;
  };

  wrapped.producer = producer;
  wrapped.consumer = consumer;

  Producers.push(producer);
  Consumers.push(consumer);
  return wrapped;
}

module.exports = {
  configKeys: configKeys,
  Util: Util,
  Producer: Producer,
  Consumer: Consumer,
  varsnap: varsnap,
  version: version,
};
